"""This is the ansys-sphinx-theme module."""
import logging
import os
import pathlib
from typing import Any, Dict

from docutils.nodes import document
from sphinx import addnodes
from sphinx.application import Sphinx

from ansys_sphinx_theme.extension.linkcode import DOMAIN_KEYS, sphinx_linkcode_resolve
from ansys_sphinx_theme.latex import generate_404  # noqa: F401

__version__ = "0.11.1"


# Declare the fundamental paths of the theme
THIS_PATH = pathlib.Path(__file__).parent.resolve()
THEME_PATH = THIS_PATH / "theme" / "ansys_sphinx_theme"
STATIC_PATH = THEME_PATH / "static"
STYLE_PATH = STATIC_PATH / "css"
JS_PATH = STATIC_PATH / "js"
CSS_PATH = STYLE_PATH / "ansys_sphinx_theme.css"
TEMPLATES_PATH = THEME_PATH / "_templates"
AUTOAPI_TEMPLATES_PATH = TEMPLATES_PATH / "autoapi"
JS_FILE = JS_PATH / "table.js"

# make logo paths available
ansys_favicon = str((STATIC_PATH / "ansys-favicon.png").absolute())
ansys_logo_black = str((STATIC_PATH / "ansys_logo_black_cropped.jpg").absolute())
ansys_logo_white = str((STATIC_PATH / "ansys_logo_white.pdf").absolute())
ansys_logo_white_cropped = str((STATIC_PATH / "ansys_logo_white_cropped.pdf").absolute())
page_404 = str((STATIC_PATH / "404.rst").absolute())
pyansys_logo_black = str((STATIC_PATH / "pyansys-logo-black-cropped.png").absolute())
pyansys_logo_white = str((STATIC_PATH / "pyansys-logo-white-cropped.png").absolute())
watermark = str((STATIC_PATH / "watermark.pdf").absolute())


def get_html_theme_path() -> pathlib.Path:
    """Return list of HTML theme paths.

    Returns
    -------
    pathlib.Path
        Path pointing to the installation directory of the theme.

    """
    return THEME_PATH.resolve()


def get_version_match(semver: str) -> str:
    """Evaluate the version match for the multi-documentation.

    Parameters
    ----------
    semver : str
        Semantic version number in the form of a string.

    Returns
    -------
    str
        Matching version number in the form of a string.

    """
    if "dev" in semver:
        return "dev"
    major, minor, *_ = semver.split(".")
    return ".".join([major, minor])


def get_autoapi_templates_dir_relative_path(path: pathlib.Path) -> str:
    """Return a string representing the relative path for autoapi templates.

    Parameters
    ----------
    path : pathlib.Path
        Path to the desired file.

    Returns
    -------
    str
        A string rerpesenting the relative path to the autoapi templates.

    """
    return os.path.relpath(
        str(AUTOAPI_TEMPLATES_PATH.absolute()), start=str(path.parent.absolute())
    )


def convert_version_to_pymeilisearch(semver: str) -> str:
    """Convert a semantic version number to pymeilisearch-compatible format.

    This function evaluates the given semantic version number and returns a
    version number that is compatible with `pymeilisearch`, where dots are
    replaced with hyphens.

    Parameters
    ----------
    semver : str
        Semantic version number in the form of a string.

    Returns
    -------
    str
        pymeilisearch-compatible version number.
    """
    version = get_version_match(semver).replace(".", "-")
    return version


def setup_default_html_theme_options(app):
    """Set up the default configuration for the HTML options.

    Parameters
    ----------
    app : ~sphinx.application.Sphinx
        Application instance for rendering the documentation.

    Notes
    -----
    This function is the only way to overwrite ``pydata-sphinx-theme``
    configuration. Variables declared in the ``theme.conf`` do not include
    inherited ones.

    """
    # Place all switchers and icons at the end of the navigation bar
    if app.config.html_theme_options.get("switcher"):
        app.config.html_theme_options.setdefault(
            "navbar_end", ["version-switcher", "theme-switcher", "navbar-icon-links"]
        )
    app.config.html_theme_options.setdefault("collapse_navigation", True)


def fix_edit_html_page_context(
    app: Sphinx, pagename: str, templatename: str, context: dict, doctree: document
) -> None:
    """Add a function that Jinja can access for returning an "edit this page" link .

    This function creates an "edit this page" link for any library.
    The link points to the corresponding file on the main branch.

    Parameters
    ----------
    app : Sphinx
        Sphinx application instance for rendering the documentation.
    pagename : str
        Name of the current page.
    templatename : str
        Name of the template being used.
    context : dict
        Context dictionary for the page.
    doctree : document
        Document tree for the page.

    Notes
    -----
    .. [1] Originally implemented by `Alex Kaszynski <https://github.com/akaszynski>`_ in
    `PyVista <https://github.com/pyvista/pyvista>`_,
    see https://github.com/pyvista/pyvista/pull/4113
    """

    def fix_edit_link_page(link: str) -> str:
        """Transform "edit on GitHub" links to the correct URL.

        This function fixes the URL for the "edit this page" link.

        Parameters
        ----------
        link : str
            Link to the GitHub edit interface.

        Returns
        -------
        str
            Link to the corresponding file on the main branch.
        """
        github_user = context.get("github_user", "")
        github_repo = context.get("github_repo", "")
        github_source = context.get("source_path", "")
        kind = context.get("github_version", "")

        if "_autosummary" in pagename:
            for obj_node in list(doctree.findall(addnodes.desc)):
                try:
                    domain = obj_node.get("domain")
                    for signode in obj_node:
                        if not isinstance(signode, addnodes.desc_signature):
                            continue
                        # Convert signode to a specified format
                        info = {}
                        for key in DOMAIN_KEYS.get(domain, []):
                            value = signode.get(key)
                            if not value:
                                value = ""
                            info[key] = value
                        if not info:
                            continue
                        # This is an API example
                        return sphinx_linkcode_resolve(
                            domain=domain,
                            info=info,
                            library=f"{github_user}/{github_repo}",
                            source_path=github_source,
                            github_version=kind,
                            edit=True,
                        )
                except ValueError as e:
                    logging.debug(f"An error occurred: {e}")  # Log the exception as debug info
                    return link

        elif pagename in ["autoapi", "api"]:
            for obj_node in list(doctree.findall(addnodes.desc)):
                domain = obj_node.get("domain")
                if domain != "py":
                    return link

                for signode in obj_node:
                    if not isinstance(signode, addnodes.desc_signature):
                        continue

                    fullname = signode["module"]
                    modname = fullname.replace(".", "/")

                    if github_source:
                        return f"http://github.com/{github_user}/{github_repo}/edit/{kind}/{github_source}/{modname}.{domain}"  # noqa: E501
                    else:
                        return f"http://github.com/{github_user}/{github_repo}/edit/{kind}/{modname}.{domain}"  # noqa: E501

        else:
            return link

    context["fix_edit_link_page"] = fix_edit_link_page


def update_footer_theme(
    app: Sphinx, pagename: str, templatename: str, context: Dict[str, Any], doctree: document
) -> None:
    """Update the version number of the Ansys Sphinx theme in the footer.

    Connect to the Sphinx application instance for rendering the documentation,
    and add the current version number of the Ansys Sphinx theme to the page context.
    This allows the theme to update the footer with the current version number.

    Parameters
    ----------
    app : ~sphinx.application.Sphinx
        Application instance for rendering the documentation.
    pagename : str
        The name of the current page.
    templatename : str
        The name of the template being used.
    context : dict
        The context dictionary for the page.
    doctree : ~docutils.nodes.document
        The document tree for the page.
    """
    context["ansys_sphinx_theme_version"] = __version__


def setup(app: Sphinx) -> Dict:
    """Connect to the sphinx theme app.

    Parameters
    ----------
    app : ~sphinx.application.Sphinx
        Application instance for rendering the documentation.

    Returns
    -------
    Dict
        Dictionary containing application status.

    """
    # Add the theme configuration
    theme_path = get_html_theme_path()
    app.add_html_theme("ansys_sphinx_theme", theme_path)
    app.config.templates_path.append(str(THEME_PATH / "components"))

    # Add default HTML configuration
    setup_default_html_theme_options(app)

    # Verify that the main CSS file exists
    if not CSS_PATH.exists():
        raise FileNotFoundError(f"Unable to locate ansys-sphinx theme at {CSS_PATH.absolute()}")
    app.add_css_file(str(CSS_PATH.relative_to(STATIC_PATH)))
    app.add_js_file(str(JS_FILE.relative_to(STATIC_PATH)))
    app.add_js_file("https://cdn.datatables.net/1.10.23/js/jquery.dataTables.min.js")
    app.add_css_file("https://cdn.datatables.net/1.10.23/css/jquery.dataTables.min.css")
    app.connect("html-page-context", update_footer_theme)
    app.connect("html-page-context", fix_edit_html_page_context)
    app.config.templates_path.append(str(TEMPLATES_PATH))
    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""This module creates the title page for a LaTeX pdf."""
from datetime import datetime
import os
from pathlib import Path

import jinja2

LATEX_SUBPKG = Path(os.path.dirname(os.path.realpath(__file__)))
COVER_TEX = LATEX_SUBPKG / "cover.tex"
PAGE_404 = LATEX_SUBPKG / "404.html"


def generate_preamble(title, watermark="watermark", date=None):
    """Generate the preamble for the PDF documentation.

    Parameters
    ----------
    title : str
        Title of the document.
    watermark : str, optional
        Name of the watermark image.
    date : ~datetime.datetime, optional
        Date of document generation. If not provided, today's date is used.

    Returns
    -------
    str
        A string representing the LaTeX source code for the preamble.

    """
    if date is None:
        date = datetime.today()
    variables = dict(_title=title, _watermark=watermark, _date=date)

    latex_jinja_env = jinja2.Environment(
        block_start_string="\BLOCK{",
        block_end_string="}",
        variable_start_string="\pyvar{",
        variable_end_string="}",
        comment_start_string="\#{",
        comment_end_string="}",
        line_statement_prefix="%%",
        line_comment_prefix="%#",
        trim_blocks=True,
        autoescape=False,
        loader=jinja2.FileSystemLoader(COVER_TEX),
    )
    template = latex_jinja_env.get_template(".")
    return template.render(variables)


def generate_404(
    owner="ansys",
    project_name="ansys-sphinx-theme",
    mail_id="pyansys.support@ansys.com",
    team_name="PyAnsys",
):
    """Generate the html body for 404 page.

    Parameters
    ----------
    owner : str, default: "ansys"
        GitHub organisation in which the project belongs to.
    project_name : str, default: "ansys-sphinx-theme"
        Name of the project.
    mail_id : str, default: "pyansys.support@ansys.com"
        E-mail address to contact.
    team_name : str, default: "PyAnsys"
       Name of the team.

    Returns
    -------
    str
        A string representing the html source code for the 404 page.

    """
    issue_page = f"https://github.com/{owner}/{project_name}/issues/"
    variables = dict(
        issue_page=issue_page, project_name=project_name, mail_id=mail_id, team_name=team_name
    )
    html_env = jinja2.Environment(loader=jinja2.FileSystemLoader(PAGE_404))
    template = html_env.get_template(".")
    return template.render(variables)



"""Module containing an example function using type hinting."""
from typing import Union


def type_hint_func(param1: int = 1, param2: str = "test", param3: Union[int, float] = 1) -> bool:
    """Summary containing the function description.

    Extended description of the function. Can span multiple lines and
    provides a general overview of the function.

    Parameters
    ----------
    param1 :
        Description of an integer parameter.
    param2 :
        Description of a string parameter.
    param3 :
        Parameter that can be either int or float using Union (typing).

    Returns
    -------
    bool
        Description of the returned value.

    Examples
    --------
    >>> func(1, 'foo', 1)
    True

    """
    return True



"""Sample functions for ansys-sphinx-theme."""


def func(arg1, arg2):
    """Summary line <should be one one line>.

    Extended description of function.  Can span multiple lines and
    provides a general overview of the function.

    .. warning::
       Use the ``.. warning::`` directive within the doc-string for any
       warnings you would like to explicitly state.  For example, if
       this method will be deprecated in the next release.

    Parameters
    ----------
    arg1 : int
        Description of arg1.
    arg2 : str
        Description of arg2.

    Returns
    -------
    bool
        Description of return value.

    Examples
    --------
    >>> func(1, 'foo')
    True

    """
    return True



"""A module containing the an extension for creating links to original source code files."""
import inspect
import os
import os.path as op
import sys

from docutils import nodes
from docutils.nodes import Node
import sphinx
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.locale import _

"""
DOMAIN_KEYS: Dictionary containing information about supported
programming languages and their corresponding domain keys.

This dictionary maps programming language codes to a list of domain keys that
can be used to extract specific information related to the language.

- 'py': Python language information keys:
  - 'module': Key to access module information.
  - 'fullname': Key to access full name information.

- 'c': C language information keys:
  - 'names': Key to access name information.

- 'cpp': C++ language information keys:
  - 'names': Key to access name information.

- 'js': JavaScript language information keys:
  - 'object': Key to access object information.
  - 'fullname': Key to access full name information.

"""
DOMAIN_KEYS = {
    "py": ["module", "fullname"],
    "c": ["names"],
    "cpp": ["names"],
    "js": ["object", "fullname"],
}


def sphinx_linkcode_resolve(
    domain: str, info: dict, library: str, source_path: str, github_version: str, edit: bool = False
) -> str or None:
    """
    Resolve the URL corresponding to a Python object for linking to the source code.

    Parameters
    ----------
    domain : str
        Domain of the object (e.g., 'py' for Python).
    info : dict
        Dictionary containing the information about the object.
        It must have the keys 'module' and 'fullname'.
    library : str
        Repository/library name where the source code is hosted.
        For example, 'ansys/ansys-sphinx-theme'.
    source_path : str
        Relative path of the source code file within the repository.
        For example, 'src'.
    github_version : str
        Version of the library in github. It can be a specific version like
        'main' or 'release/branch'.For versioned links, the version will be used in the URL.
    edit : bool, default : False
        If ``True``, the link should point to the edit page for the source code.
        Otherwise, it will point to the view page.

    Returns
    -------
    str or None
        Code URL or ``None`` if it cannot be determined.

    Notes
    -----
    This function is used by the `sphinx.ext.linkcode` extension to create the "[Source]"
    button whose link is edited in this function.

    Adapted from the link_code extension implemented in pyvista. See
    GitHub pull request: https://github.com/pyvista/pyvista/pull/4113
    (Author: Alex kaszynski <https://github.com/akaszynski>)
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    # Little clean up to avoid library.library
    if fullname.startswith(modname):
        fullname = fullname[len(modname) + 1 :]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    # Deal with our decorators properly
    while hasattr(obj, "fget"):
        obj = obj.fget

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:  # pragma: no cover
        fn = None

    if not fn:  # pragma: no cover
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            return None
        return None

    fn = op.relpath(fn, start=os.path.dirname(os.path.abspath(__file__)))
    fn_components = op.normpath(fn).split(os.sep)

    if not source_path:
        module = modname.split(".")[0]
        repo_index = fn_components.index(module)
    else:
        if source_path in fn_components:
            repo_index = fn_components.index(source_path)
        else:
            module = modname.split(".")[0]
            repo_index = fn_components.index(module)
            fn_components.insert(repo_index, source_path)
    fn = "/".join(fn_components[repo_index:])

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:  # pragma: no cover
        lineno = None

    if lineno and not edit:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    blob_or_edit = "edit" if edit else "blob"
    return f"http://github.com/{library}/{blob_or_edit}/{github_version}/{fn}{linespec}"


def link_code(app: Sphinx, doctree: Node):
    """
    Automatically add "View Source" links to the documentation.

    This function is an event handler that automatically adds "View Source" links to the
    documentation for Python, C, C++, and JavaScript objects. It is designed to work with
    the Sphinx documentation tool.

    Parameters
    ----------
    app : sphinx.application.Sphinx
        Sphinx application instance.
    doctree : docutils.nodes.Node
        Document tree for the current page.

    Notes
    -----
    The function uses the `sphinx_linkcode_resolve` function to resolve the link for each
    documented object. The link is added as a "View Source" link to the documentation
    when rendered.

    For Python, the function extracts information such as module name and fullname from
    the signature node to create the link.

    The `link_code_library` configuration option can be used to specify a custom library
    for link resolution.

    Notes
    -----
    Adapted from the original `linkcode` extension in Sphinx, see
    https://github.com/sphinx-doc/sphinx/blob/main/sphinx/ext/linkcode.py
    """
    env = app.builder.env
    html_context = getattr(env.config, "html_context", {})
    github_user = html_context.get("github_user", "")
    github_repo = html_context.get("github_repo", "")
    source_path = html_context.get("source_path", "")
    github_version = html_context.get("github_version", "main")

    if getattr(env.config, "link_code_library"):
        library = getattr(env.config, "link_code_library")

    elif github_user and github_repo:
        library = f"{github_user}/{github_repo}"

    else:
        raise AttributeError("The library should have either html_context or link_code_library.")

    github_source = (
        env.config.link_code_source if getattr(env.config, "link_code_source") else source_path
    )
    github_version = (
        env.config.link_code_branch if getattr(env.config, "link_code_branch") else github_version
    )

    if not library:
        raise AttributeError(
            "The conf.py file either should have html_context or link_code_library with github_repo and github_user."  # noqa: E501
        )

    for objnode in list(doctree.findall(addnodes.desc)):
        domain = objnode.get("domain")
        uris: set[str] = set()
        for signode in objnode:
            if not isinstance(signode, addnodes.desc_signature):
                continue

            # Convert signode to a specified format
            info = {}
            for key in DOMAIN_KEYS.get(domain, []):
                value = signode.get(key)
                if not value:
                    value = ""
                info[key] = value
            if not info:
                continue

            # Call user code to resolve the link
            uri = sphinx_linkcode_resolve(
                domain=domain,
                info=info,
                library=library,
                source_path=github_source,
                github_version=github_version,
            )
            if not uri:
                # no source
                continue

            if uri in uris or not uri:
                # only one link per name, please
                continue
            uris.add(uri)

            inline = nodes.inline("", _("[source]"), classes=["viewcode-link"])
            onlynode = addnodes.only(expr="html")
            onlynode += nodes.reference("", "", inline, internal=False, refuri=uri)
            signode += onlynode


def setup(app: Sphinx):
    """
    Initialize the linkcode extension for Sphinx.

    This function initializes the linkcode extension for Sphinx. It connects the
    `link_code` function to the "doctree-read" event and adds the "link_code_library"
    configuration option.

    Parameters
    ----------
    app : sphinx.application.Sphinx
        Sphinx application instance.

    Returns
    -------
    dict
        Dictionary containing configuration values for the extension.

    Notes
    -----
    Sphinx GitHub repository for `linkcode` extension:
    https://github.com/sphinx-doc/sphinx/blob/main/sphinx/ext/linkcode.py
    """
    app.connect("doctree-read", link_code)
    app.add_config_value("link_code_library", None, "")
    app.add_config_value("link_code_source", None, "")
    app.add_config_value("link_code_branch", None, "")
    return {"version": sphinx.__display_version__, "parallel_read_safe": True}
